package org.maachang.dbm.engine ;

import java.io.IOException;

import org.maachang.util.FileUtil;

/**
 * セクター管理オブジェクト.
 * 
 * @version 2008/01/17
 * @author masahito suzuki
 * @since MaachangDBM 1.00
 */
public class MSector {
    
    // データ長を、環境変数で変更.
    static {
        // オフセット値を計算.
        int of = ( ( MDbmEnv.ONE_SECTOR_BY_SECTOR_SIZE / 32 ) +
            ( ( ( MDbmEnv.ONE_SECTOR_BY_SECTOR_SIZE % 32 ) != 0 ) ? 1 : 0 ) ) * 4 ;
        of += 12 ;
        
        // セクタサイズを計算.
        int[] vl = MDbmEngineUtil.getSector( MDbmEnv.SECTOR ) ;
        
        ONE_DATA = vl[0]+1 ;
        SECTOR = vl[0] ;
        MASK = ~vl[0] ;
        SHIFT = vl[1] ;
        OFFSET = of ;
    }
    
    /**
     * １データ長.
     */
    protected static final int ONE_DATA ;
    
    /**
     * オフセット値.
     */
    private static final int OFFSET ;
    
    /**
     * セクタ長取得用.
     */
    private static final int SECTOR ;
    private static final int MASK ;
    private static final int SHIFT ;
    
    /**
     * Valueセクタヘッダ長.
     */
    protected static final int ONE_SECTOR_HEADER_LENGTH = 14 ;
    
    /**
     * 1セクター数.
     */
    protected static final int ONE_SECTOR = ONE_DATA+ONE_SECTOR_HEADER_LENGTH ;
    
    /**
     * １つのファイルで管理するセクター数.
     */
    protected static final int MAX_SECTOR = MDbmEnv.ONE_SECTOR_BY_SECTOR_SIZE ;
    
    /**
     * キー管理用ファイル.
     */
    private RandIO fp = null ;
    
    /**
     * 空き情報管理用.
     */
    private FlagMapFile space = null ;
    
    /**
     * 番号.
     */
    private int fileNo = -1 ;
    
    /**
     * 同期オブジェクト.
     */
    private final Object sync = new Object() ;
    
    /**
     * コンストラクタ.
     */
    private MSector() {
        
    }
    
    /**
     * コンストラクタ.
     * <BR><BR>
     * 条件を指定して、オブジェクトを生成します.
     * <BR>
     * @param fileNo 対象のファイルNoを設定します.
     * @param filename 対象のファイル名を設定します.
     * @exception Exception 例外.
     */
    public MSector( int fileNo,String filename ) throws Exception {
        if( fileNo <= -1 || fileNo >= MKey.MAX_VALUE_FILE_SIZE ||
            filename == null || ( filename = filename.trim() ).length() <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        filename = FileUtil.getFullPath( filename ) ;
        boolean isFile = FileUtil.isFileExists( filename ) ;
        RandIO fp = null ;
        if( isFile == false ) {
            //fp = new RandIO( filename,OFFSET + ( ONE_SECTOR * MAX_SECTOR ) ) ;
            fp = new RandIO( ONE_SECTOR,filename,OFFSET + ( ONE_SECTOR * MAX_SECTOR ) ) ;
        }
        else {
            //fp = new RandIO( filename ) ;
            fp = new RandIO( ONE_SECTOR,filename ) ;
        }
        this.space = initSpace( isFile,fp ) ;
        this.fp = fp ;
        this.fileNo = fileNo ;
    }
    
    /**
     * デストラクタ.
     * <BR>
     * @exception Exception 例外.
     */
    protected void finalize() throws Exception {
        this.destroy() ;
    }
    
    /**
     * オブジェクト破棄.
     */
    public void destroy() {
        synchronized( sync ) {
            if( this.space != null ) {
                this.space.clear() ;
            }
            if( this.fp != null ) {
                this.fp.destroy() ;
            }
            this.fp = null ;
            this.space = null ;
            this.fileNo = -1 ;
        }
    }
    
    /**
     * オブジェクト更新.
     * <BR><BR>
     * オブジェクトを更新します.
     * <BR>
     * @exception Exception 例外.
     */
    public void flush() throws Exception {
        synchronized( sync ) {
            if( check() == false ) {
                throw new IOException( "オブジェクトは既に破棄されています" ) ;
            }
            space.flush() ;
        }
    }
    
    private static final FlagMapFile initSpace( boolean isFile,RandIO fp ) throws Exception {
        return new FlagMapFile( isFile == false,fp.getRandomAccessFile(),0,MAX_SECTOR ) ;
    }
    
    /**
     * 新しいセクターに予約.
     * <BR><BR>
     * 新しいセクターに予約します.
     * <BR>
     * @param sector 対象のセクター情報を設定します.
     * @return boolean [true]の場合、予約できました.
     * @exception Exception 例外.
     */
    public boolean add( ValueSector sector ) throws Exception {
        if( sector == null ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        synchronized( sync ) {
            if( check() == false ) {
                throw new IOException( "オブジェクトは既に破棄されています" ) ;
            }
            int pos = space.usePos( 0 ) ;
            if( pos <= -1 ) {
                return false ;
            }
            sector.setPosition( pos ) ;
            sector.setFileNo( this.fileNo ) ;
            space.setPos( pos ) ;
        }
        return true ;
    }
    
    /**
     * 既に存在するセクター位置に情報を書き込む.
     * <BR><BR>
     * 既に存在するセクター位置に情報を書き込みます.
     * <BR>
     * @param no セクター項番を設定します.
     * @param sector 対象のセクター情報を設定します.
     * @exception Exception 例外.
     */
    public void put( int no,ValueSector sector ) throws Exception {
        if( sector == null || sector.getLength() <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        synchronized( sync ) {
            if( check() == false ) {
                throw new IOException( "オブジェクトは既に破棄されています" ) ;
            }
            if( space.getPos( no ) == false ) {
                throw new IOException( "指定セクター("+no+")は[OFF]利用されていません" ) ;
            }
            saveBySector( no,sector ) ;
        }
    }
    
    /**
     * 指定セクター項番の内容を破棄.
     * <BR><BR>
     * 指定セクター項番の内容を破棄します.
     * <BR>
     * @param no セクター項番を設定します.
     * @return ValueSector 対象のセクター内容が返されます.
     * @exception Exception 例外.
     */
    public ValueSector remove( int no ) throws Exception {
        ValueSector ret = null ;
        synchronized( sync ) {
            if( check() == false ) {
                throw new IOException( "オブジェクトは既に破棄されています" ) ;
            }
            if( space.getPos( no ) == true ) {
                ret = posBySector( no ) ;
                if( ret != null ) {
                    //removeBySector( no ) ;
                    space.removePos( no ) ;
                }
            }
        }
        return ret ;
    }
    
    /**
     * 指定セクター項番の内容を取得.
     * <BR><BR>
     * 指定セクター項番の内容を取得します.
     * <BR>
     * @param no セクター項番を設定します.
     * @return ValueSector 対象のセクター内容が返されます.
     * @exception Exception 例外.
     */
    public ValueSector get( int no ) throws Exception {
        ValueSector ret = null ;
        synchronized( sync ) {
            if( check() == false ) {
                throw new IOException( "オブジェクトは既に破棄されています" ) ;
            }
            if( space.getPos( no ) == true ) {
                ret = posBySector( no ) ;
            }
        }
        return ret ;
    }
    
    /**
     * 指定セクタ項番のヘッダ情報を取得.
     * <BR><BR>
     * 指定セクタ項番のヘッダ情報を取得します.
     * <BR>
     * @param no セクタ項番を設定します.
     * @return ValueSector 対象のヘッダセクタ内容が返されます.
     * @exception Exception 例外.
     */
    public ValueSector getHeader( int no ) throws Exception {
        ValueSector ret = null ;
        synchronized( sync ) {
            if( check() == false ) {
                throw new IOException( "オブジェクトは既に破棄されています" ) ;
            }
            if( space.getPos( no ) == true ) {
                ret = posByHeaderSector( no ) ;
            }
        }
        return ret ;
    }
    
    /**
     * ファイル項番を取得.
     * <BR><BR>
     * ファイル項番を取得します.
     * <BR>
     * @return int ファイル項番が返されます.
     */
    public int getFileNo() {
        int ret = -1 ;
        synchronized( sync ) {
            if( check() == false ) {
                ret = -1 ;
            }
            else {
                ret = this.fileNo ;
            }
        }
        return ret ;
    }
    
    /**
     * ファイル名を取得.
     * <BR><BR>
     * ファイル名を取得します.
     * <BR>
     * @return String ファイル名が返されます.
     */
    public String getFileName() {
        String ret = null ;
        synchronized( sync ) {
            if( check() == false ) {
                ret = null ;
            }
            else {
                ret = this.fp.getName() ;
            }
        }
        return ret ;
    }
    
    /**
     * 現在のデータ数を取得.
     * <BR><BR>
     * 現在のデータ数が返されます.
     * <BR>
     * @return int 現在のデータ数が返されます.
     * throws Exception 例外.
     */
    public int size() throws Exception {
        int ret = 0 ;
        synchronized( sync ) {
            if( check() == false ) {
                ret = -1 ;
            }
            else {
                ret = space.size() ;
            }
        }
        return ret ;
    }
    
    /**
     * セクター空き容量を取得.
     * <BR><BR>
     * セクター空き容量を取得します.
     * <BR>
     * @return int セクター空き容量が返されます.
     * @exception Exception 例外.
     */
    public int useSector() throws Exception {
        int ret = 0 ;
        synchronized( sync ) {
            if( check() == false ) {
                ret = -1 ;
            }
            else {
                ret = space.maxSize() - space.size() ;
            }
        }
        return ret ;
    }
    
    /**
     * このオブジェクトが有効かチェック.
     * <BR><BR>
     * このオブジェクトが有効であるかチェックします.
     * <BR>
     * @return boolean [true]の場合、有効です.
     */
    public boolean isUse() {
        boolean ret = false ;
        synchronized( sync ) {
            ret = check() ;
        }
        return ret ;
    }
    
    /**
     * １つのセクター管理ファイルサイズを取得.
     */
    protected static final long oneSectorFileSize() {
        return ( long )( OFFSET + ( ONE_SECTOR * MAX_SECTOR ) ) ;
    }
    
    /**
     * 指定データ長をセクター長に変換.
     */
    protected static final int sectorLength( int len ) {
        return ( ( len & MASK ) >> SHIFT ) + ( ( ( len & SECTOR ) != 0 ) ? 1 : 0 ) ;
    }
    
    
    private boolean check() {
        if( fp == null || space == null ) {
            return false ;
        }
        return true ;
    }
    
    private static final long getPosByAddress( int pos ) {
        return ( long )OFFSET + ( ( long )pos * ( long )ONE_SECTOR ) ;
    }
    
    private ValueSector posByHeaderSector( int pos ) throws Exception {
        byte[] b = new byte[ ValueSector.HEADER_LENGTH ] ;
        long addr = getPosByAddress( pos ) ;
        synchronized( sync ) {
            fp.read( b,addr ) ;
        }
        ValueSector ret = ValueSector.load( b ) ;
        ret.setPosition( pos ) ;
        ret.setFileNo( this.fileNo ) ;
        return ret ;
    }
    
    private ValueSector posBySector( int pos ) throws Exception {
        if( ONE_DATA <= 512 ) {
            byte[] b = new byte[ ONE_SECTOR ] ;
            long addr = getPosByAddress( pos ) ;
            synchronized( sync ) {
                fp.read( b,addr ) ;
            }
            ValueSector ret = ValueSector.load( b ) ;
            ret.setPosition( pos ) ;
            ret.setFileNo( this.fileNo ) ;
            return ret ;
        }
        else {
            byte[] b = new byte[ ValueSector.HEADER_LENGTH ] ;
            long addr = getPosByAddress( pos ) ;
            synchronized( sync ) {
                fp.read( b,addr,0,ValueSector.HEADER_LENGTH ) ;
            }
            ValueSector ret = ValueSector.load( b ) ;
            int len = ret.getLength() ;
            if( len > 0 ) {
                b = ret.getData() ;
                synchronized( sync ) {
                    fp.read( b,addr+( long )ValueSector.HEADER_LENGTH,0,len ) ;
                }
            }
            ret.setPosition( pos ) ;
            ret.setFileNo( this.fileNo ) ;
            return ret ;
        }
    }
    
    private void saveBySector( int pos,ValueSector sector ) throws Exception {
        if( sector == null ) {
            throw new IOException( "不正なセクタ["+getPosByAddress( pos )+
                "]保存を行おうとしました" ) ;
        }
        byte[] b = sector.save() ;
        long addr = getPosByAddress( pos ) ;
        int len = sector.getLength() ;
        if( len <= 0 ) {
            len = ONE_SECTOR_HEADER_LENGTH ;
        }
        else {
            len = len + ONE_SECTOR_HEADER_LENGTH ;
        }
        synchronized( sync ) {
            fp.write( b,addr,0,len ) ;
        }
        sector.setPosition( pos ) ;
        sector.setFileNo( this.fileNo ) ;
    }
    
    /*private void removeBySector( int pos ) throws Exception {
        long addr = getPosByAddress( pos ) ;
        byte[] b = new byte[ ONE_SECTOR ] ;
        Arrays.fill( b,( byte )-1 ) ;
        synchronized( sync ) {
            fp.write( b,addr ) ;
        }
    }*/
}

